"""Model setup module."""

__author__ = "Lukas Leufen, Felix Kleinert"
__date__ = '2019-12-02'

import logging
import os
import re
from dill.source import getsource

import tensorflow.keras as keras
import pandas as pd
import tensorflow as tf
import numpy as np

from mlair.model_modules.keras_extensions import HistoryAdvanced, EpoTimingCallback, CallbackHandler
from mlair.run_modules.run_environment import RunEnvironment
from mlair.configuration import path_config


class ModelSetup(RunEnvironment):
    """
    Set up the model.

    Schedule of model setup:
        #. set channels (from variables dimension)
        #. build imported model
        #. plot model architecture
        #. load weights if enabled (e.g. to resume a training)
        #. set callbacks and checkpoint
        #. compile model

    Required objects [scope] from data store:
        * `experiment_path` [.]
        * `experiment_name` [.]
        * `train_model` [.]
        * `create_new_model` [.]
        * `generator` [train]
        * `model_class` [.]

    Optional objects
        * `lr_decay` [model]

    Sets
        * `channels` [model]
        * `model` [model]
        * `hist` [model]
        * `callbacks` [model]
        * `model_name` [model]
        * all settings from model class like `dropout_rate`, `initial_lr`, and `optimizer` [model]

    Creates
        * plot of model architecture `<model_name>.pdf`

    """

    def __init__(self):
        """Initialise and run model setup."""
        super().__init__()
        self.model = None
        exp_name = self.data_store.get("experiment_name")
        self.path = self.data_store.get("model_path")
        self.scope = "model"
        path = os.path.join(self.path, f"{exp_name}_%s")
        self.model_name = path % "%s.h5"
        self.checkpoint_name = path % "model-best.h5"
        self.callbacks_name = path % "model-best-callbacks-%s.pickle"
        self._train_model = self.data_store.get("train_model")
        self._create_new_model = self.data_store.get("create_new_model")
        self._run()

    def _run(self):

        # set channels depending on inputs
        self._set_shapes()

        # build model graph using settings from my_model_settings()
        self.build_model()

        # broadcast custom objects
        self.broadcast_custom_objects()

        # plot model structure
        self.plot_model()

        # load weights if no training shall be performed
        if not self._train_model and not self._create_new_model:
            self.load_model()

        # create checkpoint
        self._set_callbacks()

        # compile model
        self.compile_model()

        # report settings
        self.report_model()

    def _set_shapes(self):
        """Set input and output shapes from train collection."""
        shape = list(map(lambda x: x.shape[1:], self.data_store.get("data_collection", "train")[0].get_X()))
        self.data_store.set("input_shape", shape, self.scope)
        shape = list(map(lambda y: y.shape[1:], self.data_store.get("data_collection", "train")[0].get_Y()))
        self.data_store.set("output_shape", shape, self.scope)

    def _set_num_of_training_samples(self):
        """ Set number of training samples - needed for example for Bayesian NNs"""
        samples = 0
        upsampling = self.data_store.create_args_dict(["upsampling"], "train")
        for data in self.data_store.get("data_collection", "train"):
            length = data.__len__(**upsampling)
            samples += length
        return samples

    def compile_model(self):
        """
        Compiles the keras model. Compile options are mandatory and have to be set by implementing set_compile() method
        in child class of AbstractModelClass.
        """
        compile_options = self.model.compile_options
        self.model.compile(**compile_options)
        self.data_store.set("model", self.model, self.scope)

    def _set_callbacks(self):
        """
        Set all callbacks for the training phase.

        Add all callbacks with the .add_callback statement. Finally, the advanced model checkpoint is added.
        """
        # create callback handler
        callbacks = CallbackHandler()

        # add callback: learning rate
        lr = self.data_store.get_default("lr_decay", scope=self.scope, default=None)
        if lr is not None:
            callbacks.add_callback(lr, self.callbacks_name % "lr", "lr")

        # add callback: advanced history
        hist = HistoryAdvanced()
        self.data_store.set("hist", hist, scope="model")
        callbacks.add_callback(hist, self.callbacks_name % "hist", "hist")

        # add callback: epo timing
        epo_timing = EpoTimingCallback()
        self.data_store.set("epo_timing", epo_timing, scope="model")
        callbacks.add_callback(epo_timing, self.callbacks_name % "epo_timing", "epo_timing")

        # add callback: early stopping
        patience = self.data_store.get_default("early_stopping_epochs", default=np.inf)
        restore_best_weights = self.data_store.get_default("restore_best_model_weights", default=True)
        assert bool(isinstance(patience, int) or np.isinf(patience)) is True
        cb = tf.keras.callbacks.EarlyStopping(patience=patience, restore_best_weights=restore_best_weights)
        callbacks.add_callback(cb, self.callbacks_name % "early_stopping", "early_stopping")

        # create model checkpoint
        callbacks.create_model_checkpoint(filepath=self.checkpoint_name, verbose=1, monitor='val_loss',
                                          save_best_only=True, mode='auto', restore_best_weights=restore_best_weights)

        # store callbacks
        self.data_store.set("callbacks", callbacks, self.scope)

    def load_model(self):
        """Try to load model from disk or skip if not possible."""
        try:
            self.model.load_model(self.model_name)
            logging.info(f"reload model {self.model_name} from disk ...")
        except OSError:
            logging.info('no local model to load...')

    def build_model(self):
        """Build model using input and output shapes from data store."""
        model = self.data_store.get("model_class")
        args_list = model.requirements()
        if "num_of_training_samples" in args_list:
            num_of_training_samples = self._set_num_of_training_samples()
            self.data_store.set("num_of_training_samples", num_of_training_samples, scope=self.scope)
            logging.info(f"Store number of training samples ({num_of_training_samples}) in data_store: "
                         f"self.data_store.set('num_of_training_samples', {num_of_training_samples}, scope="
                         f"'{self.scope}')")
        args = self.data_store.create_args_dict(args_list, self.scope)
        self.model = model(**args)
        self.get_model_settings()

    def broadcast_custom_objects(self):
        """
        Broadcast custom objects to keras utils.

        This method is very important, because it adds the model's custom objects to the keras utils. By doing so, all
        custom objects can be treated as standard keras modules. Therefore, problems related to model or callback
        loading are solved.
        """
        keras.utils.get_custom_objects().update(self.model.custom_objects)

    def get_model_settings(self):
        """Load all model settings and store in data store."""
        model_settings = self.model.get_settings()
        self.data_store.set_from_dict(model_settings, self.scope, log=True)
        self.model_name = self.model_name % self.data_store.get_default("model_name", self.scope, "my_model")
        self.data_store.set("model_name", self.model_name, self.scope)

    def plot_model(self):  # pragma: no cover
        """Plot model architecture as `<model_name>.pdf`."""
        try:
            with tf.device("/cpu:0"):
                file_name = f"{self.model_name.rsplit('.', 1)[0]}.pdf"
                keras.utils.plot_model(self.model, to_file=file_name, show_shapes=True, show_layer_names=True)
        except Exception as e:
            logging.info(f"Can not plot model due to: {e}")

    def report_model(self):
        # report model settings
        _f = self._clean_name
        model_settings = self.model.get_settings()
        model_settings.update(self.model.compile_options)
        model_settings.update(self.model.optimizer.get_config())
        df = pd.DataFrame(columns=["model setting"])
        for k, v in model_settings.items():
            if v is None:
                continue
            if isinstance(v, list):
                if len(v) > 0:
                    if isinstance(v[0], dict):
                        v = ["{" + vi + "}" for vi in [",".join(f"{_f(str(uk))}:{_f(str(uv))}" for uk, uv in d.items()) for d in v]]
                    else:
                        v = ",".join(_f(str(u)) for u in v)
                else:
                    v = "[]"
            if "<" in str(v):
                v = _f(str(v))
            df.loc[k] = str(v)
        df.loc["count params"] = str(self.model.count_params())
        df.sort_index(inplace=True)
        column_format = "ll"
        path = os.path.join(self.data_store.get("experiment_path"), "latex_report")
        path_config.check_path_and_create(path)
        for p in [path, self.path]:  # log to `latex_report` and `model`
            df.to_latex(os.path.join(p, "model_settings.tex"), na_rep='---', column_format=column_format)
            df.to_markdown(open(os.path.join(p, "model_settings.md"), mode="w", encoding='utf-8'), tablefmt="github")
        # report model summary to file
        with open(os.path.join(self.path, "model_summary.txt"), "w") as fh:
            self.model.summary(print_fn=lambda x: fh.write(x + "\n"))
        # print model code to file
        with open(os.path.join(self.path, "model_code.txt"), "w") as fh:
            fh.write(getsource(self.data_store.get("model_class")))

    @staticmethod
    def _clean_name(orig_name: str):
        mod_name = re.sub(r'^{0}'.format(re.escape("<")), '', orig_name).replace("'", "").split(" ")
        mod_name = mod_name[1] if any(map(lambda x: x in mod_name[0], ["class", "function", "method"])) else mod_name
        mod_name = mod_name[0].split(".")[-1] if any(
            map(lambda x: x in mod_name[0], ["tensorflow", "keras"])) else mod_name
        mod_name = mod_name[:-1] if mod_name[-1] == ">" else "".join(mod_name)
        return mod_name.split(".")[-1] if any(map(lambda x: x in mod_name, ["tensorflow", "keras"])) else mod_name
